var cqrParser = {
    allowedMIMEs: [],
    reSearchBody: new RegExp('<search.+?<\/search>', 'ig'),
                            //matches first <search ...> ... </search>
    reSerchTag: new RegExp('<search\\s.+?>(?=[^"<]+?<|$)', 'ig'),
                            //matches all <search params="">
    reInputTag: new RegExp('<input\\s.+?>(?=[^"<]+?<|$)', 'ig'),
                            //matches all <input params="">
    reTagParams: new RegExp('\\w+\\s*?=\\s*?".*?"', 'ig'),
                            //matches all params=""
    reUserParam: /\s+?user(="")*(?=\s*[\/>])/ig,
                            //matches Mycroft <input param="" user>
    reComments: new RegExp('#.{0,}', 'igm'),
                            //removes #comments in file globally
    reSpecials: new RegExp('[\n\r\f\t]', 'igm'), 
                            //linirizes plugin file

    init: function cqrParser_init(){
        var amArr = cqrEnv.AllowedMIMETypesInOrder.split(",");
        for (var each in amArr) cqrParser.allowedMIMEs.push(cqrTools.trim(amArr[each]));
        cqrLog.debug('Parsers Factory loaded.');
    },


    ///////////////////////
    // XML parsers section
    //
    parseXML: function parseXML(filedata, itemObject, omit_inputs){
        //Shadow912: Sherlock(Mycroft)'s souceTextEncoding patch
        //Shadow912: OpenSearch/MozSearch encoding attribute in DTD patch
        //Conversion data
        //XML DTD pattern in file:
        //(W3C recommended: REC-xml-20060816/REC-xml11-20060816)
        //  BOM<?xml version="X.X" encoding="XXX" standalone="XXX" ?>
        //var regBOMUTF16BE = /^\xFE\xFF\x00</m;
        //var regBOMUTF16LE = /^\xFF\xFE<\x00/m;
        //var regBOMUTF8 = /^\xEF\xBB\xBF</m;
        //var regUTF16BE = /^\x00<\x00\?/m;
        //var regUTF16LE = /^<\x00\?\x00/m;
        var regXMLDTD = /^[^<]*<\?xml\s+version\s*=\s*[\'\"][\w\.:-]+[\'\"]\s+encoding=[\'\"]([\w\.-]+)[\'\"].*\s*\?>.*$/m;
        itemObject.srcenc = "UTF-8"; //default encoding
        //if (regBOMUTF16BE.test(filedata)){
        //  itemObject.srcenc = "UTF-16";
        //} else if (regBOMUTF16LE.test(filedata)){
        //  itemObject.srcenc = "UTF-16";
        //} else if (regUTF16BE.test(filedata)){
        //  itemObject.srcenc = "UTF-16BE";
        //} else if (regUTF16LE.test(filedata)){
        //   itemObject.srcenc = "UTF-16LE";
        //} else if (regXMLDTD.test(filedata)){
        if (regXMLDTD.test(filedata)){
            itemObject.srcenc = RegExp.$1;
        }
        try{
            var uc = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
                     createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
            uc.charset = itemObject.srcenc;
            filedata = uc.ConvertToUnicode(filedata);
        }catch(er){
            cqrLog.warn(itemObject.filepath + " cannot be converted to Unicode! Passed as is.");
        }
        //EOf patch
        try{
            //var retItem = new Object();
            var parser = new DOMParser();
            var doc = parser.parseFromString(filedata, "text/xml");
            var roottag = doc.documentElement;
            
            if ((roottag.tagName == "parserError") ||
                (roottag.namespaceURI == "http://www.mozilla.org/newlayout/xml/parsererror.xml")){
                cqrLog.err(itemObject.filepath + ' XML data malformed!', roottag.textContent);
                itemObject.IsMalformed = true;
                itemObject.Diagnosis = roottag.textContent;
                return itemObject;
            }
            if (!((roottag.namespaceURI == "http://www.mozilla.org/2006/browser/search/") ||
                 (roottag.namespaceURI == "http://geek.net.ru/en/opensource/conquery/")  ||
                 (roottag.namespaceURI == "http://a9.com/-/spec/opensearch/1.1/"))) {
                itemObject.IsMalformed = true;
                itemObject.Diagnosis = 'XML is not in ConQuery, MozillaSearch or OpenSearch namespaces!\nGot: ' + roottag.namespaceURI;
                cqrLog.err(itemObject.filepath, itemObject.Diagnosis);
                return itemObject;
            }
            
            if (roottag.tagName.toLowerCase() == 'conqueryplugin'){
                return cqrParser.xmlConQuery(roottag, itemObject, 'ConQuery');
            } else if (roottag.tagName.toLowerCase() == 'searchplugin'){
                return cqrParser.xmlMozilla(roottag, itemObject, 'MozillaSearch');
            } else if (roottag.tagName.toLowerCase() == 'opensearchdescription'){
                return cqrParser.xmlOpenSearch(roottag, itemObject, 'OpenSearch');
            } else {
                itemObject.IsMalformed = true;
                itemObject.Diagnosis = 'Unknown root tag, got: ' + roottag.tagName;
                cqrLog.err(itemObject.filepath, itemObject.Diagnosis);
                return itemObject;
            }
        } catch (err) {
            itemObject.IsMalformed = true;
            itemObject.Diagnosis = 'cqrParser.parseXML cant parse plugin!\n' + err;
            cqrLog.err(filepath, itemObject.Diagnosis);
            return itemObject;
        }
    },
    
    xmlConQuery: function xmlConQuery(node, itemObject, plugType){
        itemObject.PluginType = plugType;
        itemObject.Title = cqrParser.xmlGetFirst(node, "ShortName", cqrMess.GetStringFromName('cqr.err.parse'));
        itemObject.Desc = cqrParser.xmlGetFirst(node, "Description");
        itemObject.qcharset = cqrParser.xmlGetFirst(node, "InputEncoding", "UTF-8");
        //Shadow912: patch for OpenSearch 1.1 compatible
        itemObject.rcharset = cqrParser.xmlGetFirst(node, "OutputEncoding", "UTF-8");
        itemObject.lang = cqrParser.xmlGetFirst(node, "Language", "*");
        //Eof patch
        itemObject.raw_icon = cqrParser.xmlGetFirst(node, "Image");
        if (!itemObject.raw_icon) {
            itemObject.iconpath = cqrParser.xmlGetFirst(node, "IconUpdateUrl");
        }
        itemObject.referrer = cqrParser.xmlGetFirst(node, "referrer");
        itemObject.opentarget = cqrParser.xmlGetFirst(node, "opentarget");
        itemObject.openwin = cqrParser.xmlGetFirst(node, "openwin");
        itemObject.forcein = cqrParser.xmlGetFirst(node, "forcein");
        itemObject.showif = cqrParser.xmlGetFirst(node, "showif");
        if (itemObject.showif) {
            var re = new RegExp(itemObject.showif, "ig");
            itemObject.si_compiled = re;
        }        
        itemObject.sitarget = cqrParser.xmlGetFirst(node, "showiftarget");        
        
        //Realizing valid <url type=""> tag
        //mimes = cqrParser.allowedMIMEs;
        var mimes = this.allowedMIMEs;
        var urlObjs = [];
        var urlObj = null;
        for (var i in mimes) {
            //will use first available valid mime "Url"
            urlObjs = cqrParser.xmlGetElements(node, "Url", mimes[i]);
            if (urlObjs[0]) {
                urlObj = urlObjs[0];
                break;
            }
        }
        
        if (urlObj) {
            //Ok, got one! 
            var urlAttrs = cqrParser.xmlGetElementAttrs(urlObj);
            if (urlAttrs) {
                itemObject.Action = urlAttrs["template"] || "";
                itemObject.Method = urlAttrs["method"] || "GET";
                //Shadow912: patch for OpenSearch 1.1 compatible
                itemObject.mimeType = urlAttrs["type"] || "text/html";
                itemObject.indexOffset = urlAttrs["indexoffset"] || 1;
                itemObject.pageOffset = urlAttrs["pageoffset"] || 1;
                //Eof patch
            }

            //Shadow912: convert SPC to percent encoding
            if (cqrParser.xmlGetElements(node, "CnvSpcToPEnc")) {
              itemObject.cnvSpcToPEnc = true;
            }
            //Eof patch

            //Now looking for "inputs"
            //Shadow912: patched for OpenSearch 1.1 Draft 3,
            //parameter extenion, Parameter element
            //and support mixed Param/Parameter element
            var lst_i = 0;
            //Eof patch
            var inputs = cqrParser.xmlGetElements(urlObj, 'Param');
            if (inputs.length){
                for (i=0; i<inputs.length; i++){
                    //... and it's attributes
                    var inputAttrs = cqrParser.xmlGetElementAttrs(inputs[i]);
                    if (!inputAttrs) continue;
                    inputAttrs["priority"] = i;
                    //Done! Saving it to query object
                    itemObject._StoreInput(inputAttrs);
                }
            //Shadow912: patched for OpenSearch 1.1 Draft 3,
            //parameter extenion, Parameter element
            //and support mixed Param/Parameter element
            //} else {
                lst_i = i;
            }
            var inputs2 = cqrParser.xmlGetElements(urlObj, 'Parameter');
            if (inputs2.length){
              for (i=0; i<inputs2.length; i++){
                //... and it's attributes
                var inputAttrs2 = cqrParser.xmlGetElementAttrs(inputs2[i]);
                if (!inputAttrs2) continue;
                inputAttrs2["priority"] = lst_i + i;
                //Done! Saving it to query object
                itemObject._StoreInput(inputAttrs2);
              }
              //} else {
            }
            if (!inputs.length && !inputs2.length) {
            //Eof patch
                ///no inputs
                //if this is non-conquery plugin we'll
                //parse it's template to give to the user
                //a friendly editable inputs.
                if (cqrEnv.ParseNonConQueryUrlTemplate && (plugType != 'ConQuery')){
                    if (itemObject.Action) {
                        var u = cqrParser.urlParser(itemObject.Action);
                        //Shadow912: patched for PATH_INFO format URL
                        if (u.pathinfo.length) {
                          itemObject.InputsParsedFromUrlTemplate = true;
                          itemObject.Action = u.protocol + '://' + u.domain + u.filepath + u.param;
                          for (i in u.pathinfo) {
                            itemObject._StorePathInfo(i, u.pathinfo[i]);
                          }
                        }
                        //Eof patch
                        if (u.inputs.length){
                            itemObject.InputsParsedFromUrlTemplate = true;
                            itemObject.Action = u.protocol + '://' + u.domain + u.filepath + u.param;
                            for (i in u.inputs){
                                var input = u.inputs[i];
                                input.priority = i;
                                itemObject._StoreInput(input);
                            }
                        }
                    }
                }
        //Shadow912: patched for OpenSearch 1.1 Draft 3,
        //parameter extenion, Parameter element
            }
        //Eof patch
            //Shadow912: patch for <MozParam xxx />
            //Now looking for "MozParam"
            var mozparam = cqrParser.xmlGetElements(urlObj, 'MozParam');
            if (mozparam.length){
                for (i=0; i<mozparam.length; i++){
                    //... and it's attributes
                    var mozAttrs = cqrParser.xmlGetMozElementAttrs(mozparam[i]);
                    if (!mozAttrs) continue;
                    mozAttrs["priority"] = i;
                    //Done! Saving it to query object
                    itemObject._StoreInput(mozAttrs);
                }
            }
            //EOf Patch
        } else {
            ///no valid url tag
        }
        return itemObject;
    },
    
    xmlMozilla: function(node, itemObject, plugType){
        return cqrParser.xmlConQuery(node, itemObject, plugType);
    },
    
    xmlOpenSearch: function(node, itemObject, plugType){
        return cqrParser.xmlConQuery(node, itemObject, plugType);
    },
    
    xmlGetElements: function(doc, strElName, strTypeFilterBy){
        var elts = doc.getElementsByTagName(strElName);
        //Shadow912: patch for JavaScript 1.8(Fx 3)
        //.getElementsByTagName() is more strict.
        //use .getElementsByTagNameNS() instead of .getElementsByTagNmae()
        //see BugZilla bug #206053
        //add OpenSearch parameter extension NS and ConQuery xml NS
        //NOTICE: diffrent between elts and elts[0],
        //        return value of .getElementsByTagName()
        const MozSNSURI = "http://www.mozilla.org/2006/browser/search/";
        const OpnSNSURI = "http://a9.com/-/spec/opensearch/1.1/";
        const OSPENSURI = "http://a9.com/-/spec/opensearch/extensions/parameters/1.0/";
        const CQRXNSURI = "http://geek.net.ru/en/opensource/conquery/";
        if (!elts[0]) elts = doc.getElementsByTagNameNS(MozSNSURI, strElName);
        if (!elts[0]) elts = doc.getElementsByTagNameNS(OpnSNSURI, strElName);
        if (!elts[0]) elts = doc.getElementsByTagNameNS(OSPENSURI, strElName);
        if (!elts[0]) elts = doc.getElementsByTagNameNS(CQRXNSURI, strElName);
        //Eof patch
        if (!elts) elts = doc.getElementById(strElname.toLowerCase());
        if (!elts) elts = doc.getElementById(strElname.toUpperCase());
        if (!elts) return '';
        if (!elts[0]) return '';
        if (!strTypeFilterBy) return elts;
        var elArr = [];
        for (var i=0; i<elts.length; i++){
            if (elts[i].attributes['type'].value == strTypeFilterBy) elArr.push(elts[i]);
        }
        return elArr;
    },

    xmlGetElementAttrs: function(element){
        //Returns some describing object
        //with all attributes, stored as 
        //atrribute_name:attribute_value
        //or null
        if (!element.attributes.length) return null;
        var attrs = [];
        for (var i=0; i<element.attributes.length; i++) {
            attrs[element.attributes[i].name.toLowerCase()] = element.attributes[i].value;
        }
        return attrs;
    },
    
    xmlGetFirst: function(doc, elName, defVal){
        if (!defVal) defVal = '';
        var collection = cqrParser.xmlGetElements(doc, elName);
        if (collection && collection[0]) {
            return collection[0].textContent || defVal;
        }else{
            return defVal;
        }
    },
    
    //Shadow912: patch for MozParam
    xmlGetMozElementAttrs: function(element){
        //Returns some describing object
        //with all attributes, stored as 
        //atrribute_name:attribute_value
        //or null
        //and 'value' = {MozParam:COND:(TrueVal:FalseVal)|Pref}
        //COND = defaultEngine|pref
        if (element.attributes.length < 2) return null;
        var attrs = [];
        attrs[element.attributes[0].name.toLowerCase()] =
          element.attributes[0].value;
        if (element.attributes[1].name.toLowerCase() == "condition"){
          switch (element.attributes[1].value){
          case "defaultEngine":
            if (element.attributes.length < 4) return null; //ad hoc check
            attrs['value'] = "{MozParam:defaultEngine:" +
              element.attributes[2].value + ":" + //trueValue
              element.attributes[3].value + "}";  //falseValue
            break;
          case "pref":
            if (element.attributes.length < 3) return null;
            attrs['value'] = "{MozParam:pref:" +
              element.attributes[2].value + "}"; //Pref
            break;
          }
        }else{
          return null;
        }
        return attrs;
    },
    //EOf patch
    
    ///////////////////////
    // SRC parsers section
    //
    parseMycroft: function parseMycroft(filedata, itemObject, omit_inputs){
        //var retItem = new cqrProtoQuery();
        if (!filedata) {
            itemObject.IsMalformed = true;
            itemObject.Diagnosis = 'cqrParser.parseMycroft got null file data';
            cqrLog.err('cqrParser.parseMycroft got null file data', filepath);
            return itemObject;
        }
        filedata = cqrParser.cleanupMycroft(filedata);
        
        if (itemObject.ext == ".cqr") {
            itemObject.PluginType = 'ConQuery';
        }else{
            itemObject.PluginType = 'Mycroft';
        }
        //SearchTagData = filedata.match(cqrParser.reSearchBody)[0];
        var mcSearch = cqrParser.getTags(cqrParser.reSerchTag, filedata);
        //Shadow912: Sherlock(Mycroft)'s souceTextEncoding patch
        //Conversion data(in fact, need sourceTextEncoding ONLY)
        var enctbl = {"0":"UTF-8", "1":"Shift_JIS", "2":"Big5",
                      "3":"EUC-KR", "25":"GB2312"}; //"0":"US-ASCII"
        if (mcSearch[0].sourcetextencoding) {
          if (!(itemObject.srcenc = enctbl[mcSearch[0].sourcetextencoding])){
            itemObject.srcenc = "UTF-8"; //default encoding
          }
        } else {
          itemObject.srcenc = "UTF-8"; //default encoding
        }
        try{
            var uc = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
                     createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
            uc.charset = itemObject.srcenc;
            filedata = uc.ConvertToUnicode(filedata);
        }catch(er){
            cqrLog.warn(itemObject.filepath + " cannot be converted to Unicode! Passed as is.");
        }
        //Rescan filedata converted to Unicode
        mcSearch = cqrParser.getTags(cqrParser.reSerchTag, filedata);
        //EOf patch
        if (mcSearch[0]){
            itemObject.Title = mcSearch[0].name || cqrMess.GetStringFromName('cqr.err.parse');
            itemObject.Action = mcSearch[0].action || "http://conquery.mozdev.org/pluginsdocs.html";
            itemObject.Desc = mcSearch[0].description || "";
            itemObject.referrer = mcSearch[0].referrer || "";
            itemObject.opentarget = mcSearch[0].opentarget || "";
            itemObject.openwin = mcSearch[0].openwin || "";
            itemObject.forcein = mcSearch[0].forcein || "";
            //XXX: Leo's malformed plugins special patch:
                var p_plgname = itemObject.filepath.split(cqrFS.path_delim);
                if (p_plgname[p_plgname.length-1] == "leo.src") {
                    //because of two "name" params defined:
                    itemObject.Title = "Leo Eng<->Ger";
                    //because of malformed action:
                    itemObject.Action = "http://dict.leo.org/";
                    itemObject.Desc = cqrMess.GetStringFromName('cqr.err.leo') + itemObject.Desc;
                } else if (p_plgname[p_plgname.length-1] == "leofrn.src"){
                    //because of two name params defined:
                    itemObject.Title = "Leo Frn<->Amd";
                    //because of malformed action:
                    itemObject.Action = "http://dico.leo.org/";
                    itemObject.Desc = cqrMess.GetStringFromName('cqr.err.leo') + itemObject.Desc;
                }
            //EOf leo patch
            itemObject.Method = mcSearch[0].method || "GET";
            itemObject.qcharset = mcSearch[0].querycharset || "UTF-8";
            itemObject.showif = mcSearch[0].showif || ""; 
            if (itemObject.showif) {
                var re = new RegExp(itemObject.showif, "ig");
                itemObject.si_compiled = re;
            }
            itemObject.sitarget = mcSearch[0].showiftarget || "[:selection]";
            if (!omit_inputs) {
                var inp = cqrParser.getTags(cqrParser.reInputTag, filedata);
                for (var each in inp) itemObject._StoreInput(inp[each]);
            }
            //retItem.eId = cqrParser.composeId(filepath);
            itemObject.ShowInMenu = true;
        }else{
            itemObject.IsMalformed = true;
            itemObject.Diagnosis = 'cqrParser.parseMycroft cant find <search ...> tag';
            cqrLog.err('cqrParser.parseMycroft cant find <search ...> tag', filepath);
            return itemObject;
        }
        return itemObject; 
    },
    
    getTags: function(reTag, str){
        //This extracts all <tag param="val< - > ue"> patterns from str
        var TagsData = [];
        var TagsBuff = [];
        var cqrTagsRE = reTag;
        while (TagsBuff!=null){
            TagsBuff = cqrTagsRE.exec(str);
            if (TagsBuff!=null){
                TagsData[TagsData.length]=cqrParser.tagParams(TagsBuff[0]);
                TagsData[TagsData.length-1].priority = cqrTagsRE.lastIndex;
            }
        }
    return TagsData;
    },

    tagParams: function(tag){
      var Params = new Object;
      var ParamsBuff = "  ";
      var ParamsData = [];
      //To correct Mycroft's "user" in <input>.
      tag = String(tag).replace(cqrParser.reUserParam, ' value="[:selection]"');
      Params["value"]= cqrMess.GetStringFromName('cqr.err.parse');
      ParamsBuff = cqrParser.reTagParams.exec(tag);
      while (ParamsBuff!=null){
            if (ParamsBuff!=null){
                var ParamsArr = String(ParamsBuff[0]).split(/=\s*\"/);
                ParamsArr[1] = ParamsArr[1].replace(/^\s+/, "");
                Params[String(ParamsArr[0].replace(/\s+/,"")).toLowerCase()] = ParamsArr[1].replace(/\"\s*/, "");
                ParamsBuff = cqrParser.reTagParams.exec(tag);
            }
        }
    return Params;
    },

    cleanupMycroft: function(strData){
        strData = strData.replace(cqrParser.reComments, '');
        strData = strData.replace(cqrParser.reSpecials, '  ');
        return strData;
    },
    
    urlParser: function(strURL) {
        //strURL = 'lingvo.yandex.ru';
        //Shadow912: patch for <MozParam xxx />
        //Shadow912: notice for scope of urlObj
        //urlObj = new Object;
        var urlObj = new Object;
        //EOf patch
        urlObj.protocol = '';
        urlObj.domain = '';
        urlObj.path = '';
        
        var URLi = Components.classes["@mozilla.org/network/url-parser;1?auth=maybe"].
                       createInstance(Components.interfaces.nsIURLParser);
        var specLen = -1, schemePos = {}, schemeLen = {}, authorityPos = {}, authorityLen = {}, pathPos = {}, pathLen = {};
        URLi.parseURL(strURL, specLen, schemePos, schemeLen, authorityPos, authorityLen, pathPos, pathLen);
        //cqrLog.info(strURL);
        urlObj.protocol = strURL.substr(schemePos.value, schemeLen.value);
        urlObj.domain = strURL.substr(authorityPos.value, authorityLen.value);
        urlObj.path = strURL.substr(pathPos.value, pathLen.value);
        
        var filepathPos = {}, filepathLen = {}, paramPos = {}, paramLen = {}, queryPos = {}, queryLen = {}, refPos = {}, refLen = {};
        URLi.parsePath(urlObj.path, -1, filepathPos, filepathLen, paramPos, paramLen, queryPos, queryLen, refPos, refLen);

        urlObj.filepath = urlObj.path.substr(filepathPos.value, filepathLen.value);
        urlObj.param = urlObj.path.substr(paramPos.value, paramLen.value);
        urlObj.query = urlObj.path.substr(queryPos.value, queryLen.value);
        urlObj.ref = urlObj.path.substr(refPos.value, refLen.value);
        urlObj.inputs = [];
        //Shadow912: patched for PATH_INFO format URL
        urlObj.pathinfo = [];

        //if (!urlObj.query) return urlObj;
        if (!urlObj.query) { //Condition is adhoc...
          urlObj.filepath = "/";
          urlObj.pathinfo = urlObj.path.substr(1).split('/');
          return urlObj;
        }
        //Eof patch
        var delim = '';
        if (urlObj.query.indexOf(';')>=0) delim = ';';
        if (urlObj.query.indexOf('&')>=0) delim = '&';
        if (delim){
            var pairs = urlObj.query.split(delim);
            for (var i in pairs){
                var pair = pairs[i].split('=');
                urlObj.inputs.push({'name':pair[0], 'value':pair[1]});
            }
        } else {
            if (urlObj.query.indexOf('=')>=0) {
                //url has only single pair
                pair = urlObj.query.split('=');
                urlObj.inputs.push({'name':pair[0], 'value':pair[1]});
            } else {
                //single param with no value or malformed url
                urlObj.inputs.push({'name':urlObj.query, 'value':''});
            }
        }
        return urlObj;
    }

}

cqrParser.init();
